from cnvrg.modules.base_module import CnvrgBase
import cnvrg.helpers.apis_helper as apis_helper
from cnvrg.modules.errors import CnvrgError
import cnvrg.helpers.export_library_helper as export_library_helper
from cnvrg.modules.project import Project
from cnvrg.helpers.param_build_helper import LIBRARY, PROJECT, parse_params
from cnvrg.helpers.url_builder_helper import url_join
from cnvrg.helpers.args_helper import args_to_string, validate_args
from cnvrg.modules.experiment import Experiment
from cnvrg.helpers.param_helper import wrap_string_to_list
import os
DEFAULT_WORKING_DIR = os.path.expanduser("~/cnvrg_libraries")

class Library(CnvrgBase):
    def __init__(self, library, project=None, working_dir=None, _info=None):
        owner, library = parse_params(library, LIBRARY)
        self.__working_dir = working_dir or DEFAULT_WORKING_DIR
        try:
            self.project = Project(project)
        except CnvrgError as e:
            self.project = None

        self.__library = library
        self.__owner = apis_helper.credentials.owner
        self.__info = _info
        self.__path = None
        self.__cloned = self.__lib_loaded()

    def __base_url(self):
        return url_join("users", self.__owner, "libraries", self.__library)

    @staticmethod
    def export(*paths):
        paths = [os.path.abspath(p) for p in paths]
        resp = export_library_helper.export_library(*paths)
        if resp.get("library"):
            lib_key = resp.get("library").get("key")
            return Library(lib_key)
        else:
            raise CnvrgError(resp.get("message"))

    @staticmethod
    def list():
        owner = apis_helper.credentials.owner
        lib_list = list(map(lambda x: Library(url_join(x.get('owner'), x.get("title")), _info=x), apis_helper.get(url_join("users", owner, "libraries")).get("libraries")))
        return {l.title(): l for l in lib_list}



    def info(self, force=False):
        """
        get info about this library
        :param force: to force api fetching
        :return: dict represent the library
        {clone_cmd: string,
        command: string,
        arguments: list of key: values(list),
        number_of_experiments: integer,
        description: string}
        """
        if self.__info and not force:
            return self.__info
        self.__info = self.__fetch_info()
        return self.info()


    def __fetch_info(self):
        try:
            resp = apis_helper.get(self.__base_url())
            return resp.get("library")
        except Exception as e:
            raise CnvrgError("Cant find library")



    def load(self):
        """
        load library to your local directory
        :param working_dir: path to clone the library to
        """
        info = self.info()
        lib_dir = self.__lib_path()
        os.makedirs(lib_dir, exist_ok=True)
        export_library_helper.load_library(info.get("package"), lib_dir)
        self.__path = lib_dir
        return lib_dir

    def __default_args(self):
        return {
            "project_dir": os.path.abspath(self.project.get_working_dir()),
            "output_dir": os.path.abspath(self.project.get_output_dir())
        }


    def arguments(self):
        return self.info().get("exp_arguments")

    def __mix_args(self, args=None, with_many=True, with_defaults=True):
        args = args or {}
        #exp arguments are key: values
        library_args = self.info().get("exp_arguments") or {}
        all_args = self.__default_args() if with_defaults else {}
        all_args = {
            **all_args,
            **library_args,
            **args
        }
        if not with_many: validate_args(all_args)
        return all_args


    def run(self, computes=None, compute=None, datasets=None, dataset=None, image=None, commit=None, title=None, local=True, **kwargs):
        ## support
        computes = computes or wrap_string_to_list(computes) or wrap_string_to_list(compute) or []
        datasets = datasets or wrap_string_to_list(datasets) or wrap_string_to_list(dataset) or []

        if "local" in computes or len(computes) == 0: return self.__run_local(title=title, **kwargs)
        if not self.project: raise CnvrgError("You should specify Project on the library constructor")
        args = self.__mix_args(kwargs, with_many=True, with_defaults=False)
        hyper_search_payload = self.project.run_task(cmd=self.info().get("command"), templates=computes, datasets=datasets, image=image, commit=commit, title=title, library=self.__library, arguments=args)


        return [Experiment(slug) for slug in hyper_search_payload.get("experiments")]



    def __run_local(self, title=None, **kwargs):
        if not self.project:
            raise CnvrgError("Cant run a library not in a project context, please cd into cnvrg project")
        if not self.__lib_loaded(): self.load()
        args = self.__mix_args(kwargs, with_many=False)

        ## prepare script.
        raw_script = "{cmd} {args}".format(cmd=self.info().get("command"), args=args_to_string(args))
        return [Experiment.run(raw_script, title=title, project=self.project, local=True, working_directory=self.__lib_path(), library=self.__library)]


    def __lib_loaded(self):
        return os.path.exists(os.path.join(self.__lib_path(), "library.yml"))


    def __lib_path(self):
        working_dir = self.__working_dir
        return os.path.join(working_dir, self.info().get("slug"))


    def __str__(self):
        return """{title}
        Description: {description}
        Arguments: {arguments}
        Command: {command}
        """.format(**{**self.info(), "arguments": args_to_string(self.info().get("arguments"))})

    def title(self):
        return url_join(self.info().get("owner"), self.info().get("title"))











from cnvrg.modules.base_module import CnvrgBase
from cnvrg.helpers.parallel_helper import safe_parallel
from cnvrg.helpers.status_helpers import CommitStatus
import cnvrg.helpers.files_helper as files_helper
import cnvrg.helpers.apis_helper as apis_helper
from typing import Dict
from threading import Thread
import os
import time


class CommitUploader(CnvrgBase):
    def __init__(self, storage, status: Dict):
        self.storage = storage
        self.element = storage.element
        self.sha1 = None
        self.status = status
        self.added_files = []
        self.saved_files = []
        self.finish_adding = False
        self.finish_saving = False


    def __base_url(self):
        return os.path.join(self.element.get_base_url(), "commits", self.sha1)


    def start_commit(self, force=False, new_branch=False):
        self.sha1 = self.element.start_commit(force=force, new_branch=new_branch)


    def upload_files(self):
        ## init
        delete_thread = Thread(target=self.__delete_files_thread)
        upload_thread = Thread(target=self.__upload_files_thread)
        ## start
        delete_thread.start()
        upload_thread.start()
        ## join
        delete_thread.join()
        upload_thread.join()


    def end_commit(self):
        self.element.end_commit(sha1=self.sha1)

    def __upload_files_thread(self):
        add_files = Thread(target=self.__add_files_thread)
        add_files.run()
        save_files = Thread(target=self.__save_files_thread)
        save_files.run()
        safe_parallel(func=self.storage.upload_file, list=self.__upload_files_gen, progressbar={"total": len(self.files_to_upload()), "desc": "uploading files"})
        add_files.join()
        self.finish_saving = True
        save_files.join()


    def __upload_files_gen(self, chunk_size=1000):
        while True:
            if len(self.added_files) == 0 and self.finish_adding:
                break
            to_add = self.added_files[:chunk_size]
            self.added_files = self.added_files[chunk_size:]
            yield to_add
            self.saved_files.extend(to_add)

    def files_to_upload(self):
        return self.status[CommitStatus.UPDATED_ON_LOCAL]

    def __add_files_thread(self, chunk_size=1000):
        files_to_add = self.files_to_upload()
        for i in range(0, len(files_to_add), chunk_size):
            files = files_to_add[i:i+chunk_size]
            added_files = self._add_files_call(files)
            self.added_files.extend(added_files)
        self.finish_adding = True


    def _add_files_call(self, files):
        mapped_files = list(map(self.__map_file, files))
        resp = apis_helper.post(os.path.join(self.__base_url(), 'add'), data={"files": mapped_files})
        return resp.get('blobs')

    def __map_file(self,file):
        return files_helper.get_file_props(file, self.element.working_dir)


    def __save_file_call(self, files):
        apis_helper.post(os.path.join(self.__base_url(), 'save'), data={"files": files})

    def __save_files_thread(self, chunk_size=1000):
        while True:
            if self.finish_saving and len(self.saved_files) == 0:
                break
            elif len(self.saved_files) == 0:
                time.sleep(5)
            to_save = self.saved_files[:chunk_size]
            self.__save_file_call(to_save)
            self.saved_files = self.saved_files[chunk_size:]

    def __delete_files_call(self, files):
        apis_helper.post(os.path.join(self.__base_url(), 'delete'), data={"files": files})

    def __delete_files_thread(self, chunk_size=1000):
        files_to_delete = self.status[CommitStatus.DELETED_ON_LOCAL]
        for i in range(0, len(files_to_delete), chunk_size):
            self.__delete_files_call(files_to_delete[i:i+chunk_size])








import os
import yaml
from cnvrg.helpers.apis_helper import post as apis_post, get as apis_get, credentials
from cnvrg.modules.cnvrg_files import CnvrgFiles
from cnvrg.modules.errors import CnvrgError
import cnvrg.helpers.hyper_search_helper as hyper_search_helper
import cnvrg.helpers.logger_helper as logger_helper
import cnvrg.helpers.config_helper as config_helper
import cnvrg.helpers.env_helper as env_helper
import cnvrg.helpers.param_build_helper as param_build_helper
from cnvrg.helpers.url_builder_helper import url_join

class Project(CnvrgFiles):
    def __init__(self, project=None, project_url=None, working_dir=None):
        if project_url:
            owner_slug, project_slug = Project.get_owner_and_project_from_url(project_url)
        else:
            owner_slug, project_slug = param_build_helper.parse_params(project, param_build_helper.PROJECT, working_dir=working_dir)
        self.__owner = owner_slug or credentials.owner
        self.__project = project_slug

        in_dir = config_helper.is_in_dir(config_helper.CONFIG_TYPE_PROJECT, project_slug, working_dir)
        super(Project, self).__init__()
        if in_dir:
            self._set_working_dir(config_helper.find_config_dir(path=working_dir))
            self.in_dir = True
        elif not self.__project or not self.__owner:
            raise CnvrgError("Cant init project without params and outside project directory")

    def __load_defaults(self):
        self.image = 'cnvrg'
        self.default_template = 'medium'

    def get_base_url(self):
        return "users/{owner}/projects/{project}".format(owner=self.__owner, project=self.__project)

    def _default_config(self):
        return {
            "project": self.__project,
            "owner": self.__owner,
            "commit": None
        }

    def web_url(self):
        return url_join(credentials.web_url(), self.__owner, 'projects', self.__project)

    def get_project_name(self):
        return self.__project

    def get_output_dir(self):
        return self.get_working_dir()

    def run_task(self, cmd, **kwargs):
        #if "grid" in kwargs: kwargs["grid"] = hyper_search_helper.load_hyper_search(kwargs["grid"])
        kwargs = {**env_helper.get_origin_job(), **kwargs}
        resp = apis_post(url_join(self.get_base_url(), "experiments"), data={"cmd": cmd, **kwargs})
        hyper_search = resp.get("hyper_search")
        if not hyper_search:
            raise CnvrgError("Cant run experiment")
        hyper_search_url = url_join(self.web_url(),
                                    "experiments?grid={grid}".format(grid=hyper_search.get("slug")[0:4]))
        logger_helper.log_message(
            "Grid {grid_slug} is running, follow at: {url}".format(grid_slug=hyper_search.get("slug"),
                                                                  url=hyper_search_url))
        return hyper_search

    #
    # def create_experiment(self, command, commit='latest', image=None, templates=None, datasets=None, title=None):
    #     experiment_data = {
    #         "cmd": command,
    #         "commit": commit,
    #         "image": image,
    #         "templates": templates,
    #         "datasets": datasets,
    #         "title": title,
    #     }
    #     return apis_post(url_join(self.__base_url(), "experiments"), data=experiment_data)
    #
    # def get_experiment(self, id):
    #     return apis_get(url_join(self.__base_url(),"experiments", id))
    #
    #
    #
    # def create_endpoint(self, name="testEndpoint", min_replica=1, max_replica=1, templates=None, commit="latest", function="predict", file="deploy.py"):
    #     endpoint = {
    #         "name": name,
    #         "min_replica": min_replica,
    #         "max_replica": max_replica,
    #         "templates": templates,
    #         "models": [{
    #             "commit": commit,
    #             "function":function,
    #             "file": file
    #         }]
    #     }
    #     return apis_post(url_join(self.__base_url(), "endpoints"), data=endpoint)
    #
    #
    #
    #



import click
import os
import cnvrg.actions.project_actions  as project_actions



@click.group()
@click.option("--log-level", default="INFO", help="log level")
def project(log_level):
    pass

@project.command()
@click.option('--working_dir', '-w', default=os.getcwd(), help='working dir')
@click.option('--commit', '-c', default=None, help='commit to clone')
@click.argument("project_path")
def clone(project_path, working_dir, commit):
    project_actions.clone(project_path, working_dir=working_dir, commit_sha1=commit)

@project.command()
def download():
    project_actions.download()


@project.command()
def upload():
    project_actions.upload()



from typing import List, Callable, Dict
from cnvrg.helpers.env_helper import POOL_SIZE
import types
from multiprocessing.pool import Pool
from tqdm import tqdm

def safe_parallel(func: Callable, list, pool_size: int = POOL_SIZE, progressbar: Dict=None):
    """
    safe parallel function, it support
    :param func: function to execute every time
    :param list: generator or list
    :param pool_size: how much processes will work
    :param progressbar: tqdm paramteres (if passing generator, total is mendatory (if you want progressbar)
    :return:
    """
    results = []
    pbar = None
    if not isinstance(list, types.GeneratorType):
        gen = generator_wrap(list)
    else:
        gen = list
    with Pool(processes=pool_size) as pool:
        try:
            if progressbar:
                pbar = tqdm(**progressbar)
            for batch in gen:
                for res in pool.imap_unordered(func, batch):
                    if progressbar:
                        pbar.update()
                    results.append(res)
        except Exception as e:
            raise e
        finally:
            pool.close()
        return results


def generator_wrap(llist: List, chunks_size: int=None):
    llist = list(llist)
    chunks_size = chunks_size or 10000
    for i in range(0, len(llist), chunks_size):
        yield llist[i:i+chunks_size]



import yaml
import os
import tempfile
import tarfile
import cnvrg.modules.errors as errors
import cnvrg.actions.project_actions as project_actions
import cnvrg.helpers.apis_helper as apis_helper
import cnvrg.helpers.files_helper as files_helper
import cnvrg.helpers.random_helper as random_helper

DEFAULT_WORKING_DIR = os.path.expanduser("~/cnvrg_libraries")

class LibraryConfig():
    def __init__(self, working_dir):
        working_dir = working_dir or os.curdir
        self.working_dir = working_dir
        self.config_path = os.path.join(working_dir, 'library.yml')
        if not os.path.exists(self.config_path): raise errors.CnvrgError("Cant find library in directory.")
        with open(self.config_path, 'r') as f:
            self.config_yaml = yaml.load(f)
            self.__load_fields()


    def __load_fields(self):
        self.__inflate_defaults()
        self.__verify_key_exists("command", "Please enter your library command")
        self.__inflate_file("requirements.txt", "requirements")
        self.__inflate_file("readme.me", "documentation")

    def __inflate_file(self, file, key):
        fpath = os.path.join(self.working_dir, file)
        if os.path.exists(fpath):
            self.config_yaml[key] = open(fpath, 'r').read()

    def __inflate_defaults(self):
        self.config_yaml = {**{
            "title": os.path.basename(self.working_dir),
            "description": "Cnvrg Library",
            "icon": None,
            "version": "1",
            "key": os.path.basename(self.working_dir)
        }, **self.config_yaml}

    def __verify_key_exists(self, key, prompt_message=None, alternative=None):
        if key not in self.config_yaml: self.__prompt_for_key(key, prompt_message)

    def __prompt_for_key(self,key, question=None):
        answer = input(question)
        self.config_yaml[key] = answer

    def getter(self):
        return self.config_yaml



def get_config(working_dir=None):
    return LibraryConfig(working_dir).getter()

def export_library(*paths):
    tar_path = os.path.join(os.curdir, "{title}.tar".format(title=random_helper.random_string(10)))
    with tarfile.open(tar_path, "w:gz") as tar_handle:
        for path in paths:
            for file in files_helper.expand(path, [".cnvrg"]):
                file = os.path.join(path, file)
                tar_handle.add(file, arcname=os.path.relpath(file, path))
    file = {"file": open(tar_path, "rb")}
    resp = apis_helper.send_file(apis_helper.url_join("users", apis_helper.credentials.owner, "libraries"), files=file)
    os.remove(tar_path)
    return resp


def load_library(tar_url, library_dest):
    f = apis_helper.download_raw_file(tar_url)
    with tempfile.NamedTemporaryFile() as tfile:
        tfile.write(f)
        tfile.seek(0)
        with tarfile.open(tfile.name, "r:gz") as tar_handle:
            tar_handle.extractall(library_dest)